# tests/test_clustering.py
import pytest
import numpy as np
import re

from algomake.models.clustering import KMeans, _euclidean_distance

# --- Helper function for testing inertia ---
def _calculate_inertia(X, labels, centroids):
    """Calculates the sum of squared distances of samples to their closest cluster center."""
    inertia = 0
    for i in range(len(X)):
        # Ensure labels[i] is a valid index for centroids
        if 0 <= labels[i] < len(centroids):
            inertia += np.sum((X[i] - centroids[labels[i]])**2)
        else:
            # Handle cases where a label might be out of bounds if something went wrong,
            # though this shouldn't happen with correct KMeans implementation.
            raise ValueError(f"Label {labels[i]} is out of bounds for {len(centroids)} centroids.")
    return inertia

# --- Fixtures for Synthetic Data ---

@pytest.fixture
def kmeans_data_simple():
    """Generates simple 2-cluster data."""
    np.random.seed(42)
    # Cluster 1 (around [2, 2])
    X1 = np.random.randn(50, 2) * 0.8 + np.array([2, 2])
    # Cluster 2 (around [-2, -2])
    X2 = np.random.randn(50, 2) * 0.8 + np.array([-2, -2])
    return np.vstack((X1, X2))

@pytest.fixture
def kmeans_data_three_clusters():
    """Generates data for 3 clusters."""
    np.random.seed(42)
    X1 = np.random.randn(30, 2) * 0.7 + np.array([0, 0])
    X2 = np.random.randn(30, 2) * 0.7 + np.array([5, 0])
    X3 = np.random.randn(30, 2) * 0.7 + np.array([2.5, 4])
    return np.vstack((X1, X2, X3))


# --- Tests for KMeans ---

def test_kmeans_fit_predict_simple(kmeans_data_simple):
    """Test KMeans on simple 2-cluster data."""
    X = kmeans_data_simple
    model = KMeans(n_clusters=2, random_state=42, max_iter=100)
    model.fit(X)

    labels = model.predict(X)
    assert labels.shape == (X.shape[0],)
    
    # Check if exactly 2 unique labels were found (0 and 1)
    unique_labels = np.unique(labels)
    assert len(unique_labels) == 2
    assert set(unique_labels) == {0, 1}

    # Check if centroids are reasonably placed (accounting for label permutation)
    centroids = model.cluster_centers_
    expected_centers = np.array([[2, 2], [-2, -2]]) # The true centers used to generate data

    # Calculate distances from learned centroids to expected centers
    dist0_to_exp0 = _euclidean_distance(centroids[0], expected_centers[0])
    dist0_to_exp1 = _euclidean_distance(centroids[0], expected_centers[1])
    dist1_to_exp0 = _euclidean_distance(centroids[1], expected_centers[0])
    dist1_to_exp1 = _euclidean_distance(centroids[1], expected_centers[1])

    # Case 1: Centroid 0 matches exp_center 0, Centroid 1 matches exp_center 1
    case1_match = (dist0_to_exp0 < 0.5) and (dist1_to_exp1 < 0.5)
    # Case 2: Centroid 0 matches exp_center 1, Centroid 1 matches exp_center 0 (labels swapped)
    case2_match = (dist0_to_exp1 < 0.5) and (dist1_to_exp0 < 0.5)

    assert case1_match or case2_match, \
        f"Centroids not close to expected centers. Learned: {centroids}, Expected: {expected_centers}"

    # Check inertia (should be relatively low for good clustering)
    inertia = _calculate_inertia(X, labels, centroids)
    # --- CHANGE THIS LINE ---
    assert inertia < 110 # Adjusted threshold from 100 to 110
    # print(f"KMeans Simple Inertia: {inertia}")


def test_kmeans_fit_predict_three_clusters(kmeans_data_three_clusters):
    """Test KMeans on 3-cluster data."""
    X = kmeans_data_three_clusters
    model = KMeans(n_clusters=3, random_state=42, max_iter=100)
    model.fit(X)

    labels = model.predict(X)
    assert labels.shape == (X.shape[0],)
    
    unique_labels = np.unique(labels)
    assert len(unique_labels) == 3
    assert set(unique_labels) == {0, 1, 2}

    centroids = model.cluster_centers_
    expected_centers = np.array([[0, 0], [5, 0], [2.5, 4]])
    
    # We need to find the best permutation match for centroids and expected centers
    # This is more complex for 3 clusters, so we'll rely on inertia being low and counts being balanced
    
    # Simple check for centroid count and shape
    assert centroids.shape == (3, X.shape[1])

    # Check inertia (should be relatively low for good clustering)
    inertia = _calculate_inertia(X, labels, centroids)
    assert inertia < 80 # Adjust threshold as needed
    # print(f"KMeans Three Clusters Inertia: {inertia}")

    # Optionally, for robustness, one could check if each original group
    # from the fixture has its points predominantly assigned to one cluster.
    # For instance, check points from X1 are mostly in one cluster.
    # This is more involved and might be overkill for initial tests.

def test_kmeans_initial_state():
    """Test initial state of KMeans."""
    model = KMeans()
    assert model.centroids is None
    assert model.labels_ is None
    assert model._is_fitted is False

def test_kmeans_predict_before_fit_raises_error():
    """Test that predict raises RuntimeError if not fitted."""
    model = KMeans()
    X = np.array([[1, 2]])
    expected_message = "Estimator not fitted. Call fit() before predict()."
    with pytest.raises(RuntimeError, match=re.escape(expected_message)):
        model.predict(X)

def test_kmeans_cluster_centers_before_fit_raises_error():
    """Test that cluster_centers_ property raises RuntimeError if not fitted."""
    model = KMeans()
    expected_message = "Estimator not fitted. Call fit() before predict()."
    with pytest.raises(RuntimeError, match=re.escape(expected_message)):
        _ = model.cluster_centers_

def test_kmeans_invalid_n_clusters():
    """Test that KMeans raises ValueError for invalid n_clusters."""
    with pytest.raises(ValueError, match="n_clusters must be a positive integer."):
        KMeans(n_clusters=0)
    with pytest.raises(ValueError, match="n_clusters must be a positive integer."):
        KMeans(n_clusters=-1)
    with pytest.raises(ValueError, match="n_clusters must be a positive integer."):
        KMeans(n_clusters=1.5)

def test_kmeans_invalid_max_iter():
    """Test that KMeans raises ValueError for invalid max_iter."""
    with pytest.raises(ValueError, match="max_iter must be a positive integer."):
        KMeans(max_iter=0)
    with pytest.raises(ValueError, match="max_iter must be a positive integer."):
        KMeans(max_iter=-1)
    with pytest.raises(ValueError, match="max_iter must be a positive integer."):
        KMeans(max_iter=1.5)

def test_kmeans_invalid_tol():
    """Test that KMeans raises ValueError for invalid tol."""
    with pytest.raises(ValueError, match="tol must be a non-negative number."):
        KMeans(tol=-0.001)

def test_kmeans_get_set_params():
    """Test get_params and set_params for KMeans."""
    model = KMeans(n_clusters=3, max_iter=50, tol=1e-3, random_state=10)
    params = model.get_params()
    assert params['n_clusters'] == 3
    assert params['max_iter'] == 50
    assert params['tol'] == 1e-3
    assert params['random_state'] == 10

    model.set_params(n_clusters=5, max_iter=100, tol=1e-5)
    assert model.n_clusters == 5
    assert model.max_iter == 100
    assert model.tol == 1e-5

def test_kmeans_n_samples_less_than_n_clusters_raises_error():
    """Test that KMeans raises ValueError if n_samples < n_clusters."""
    X_small = np.array([[1, 2], [3, 4]]) # 2 samples
    with pytest.raises(ValueError, match=re.escape("n_samples=2 should be >= n_clusters=3. Decrease n_clusters or add more samples.")):
        model = KMeans(n_clusters=3, random_state=42)
        model.fit(X_small)